#!/usr/bin/perl -wall

sub main()
{
	
	if(scalar(@ARGV) != 1){
		die "Usage: $0 </path/to/unafold>\n";
	}
	
	my $unafold_path = $ARGV[0];
	
	my $tstacki_dgd = "$unafold_path/data/tstacki.DGD";
	my $tstacki_dhd = "$unafold_path/data/tstacki.DHD";

	open(DGD, $tstacki_dgd) or
		die "Unable to open $tstacki_dgd\n";
	
	open(DHD, $tstacki_dhd) or
		die "Unable to open $tstacki_dhd\n";
		
	# Bases in unafold are ordered as: A C G T
	my @base = ('A', 'C', 'G', 'T');
	my $A = 0;
	my $C = 1;
	my $G = 2;
	my $T = 3;

	for(my $i = $A;$i <= $T;$i++){
		
		for(my $j = $A;$j <= $T;$j++){
		
			my $line;
			
			if( !($line = <DGD>) ){
				die "Unable to read from $tstacki_dgd\n";
			}
			
			chomp($line);
			
			my @dgd = split /\s/, $line;
			
			if( !($line = <DHD>) ){
				die "Unable to read from $tstacki_dhd\n";
			}
			
			chomp($line);
			
			my @dhd = split /\s/, $line;
			
			printf "// This C++ code is generated by extracting the values from the tstacki.DGD and tstacki.DHD\n";
			printf "// files provided by UNAfold. The extraction is performed by the script parse_tstacki.pl\n";
			
			for(my $k = $A;$k <= $T;$k++){
				
				LOOP: for(my $l = $A;$l <= $T;$l++){
					
					my $index = $k*4 + $l;
					
					if($dgd[$index] eq "inf"){
						next LOOP;
					}
					
					if($dhd[$index] eq "inf"){
						die "Unmatched inf found\n";
					}
					
					my $entropy = ($dhd[$index] - $dgd[$index])/(273.15 + 37.0);
					
					if($entropy eq "0"){
						$entropy = "0.00";
					}
					
					# The plus strand orientation
					printf "param_loop_terminal_H[$base[$i]$base[$k]][$base[$j]$base[$l]] = $dhd[$index]f;\n";
					printf "param_loop_terminal_S[$base[$i]$base[$k]][$base[$j]$base[$l]] = $entropy" . "f;\n";
					
					# The minus strand orientation
					printf "param_loop_terminal_H[$base[$l]$base[$j]][$base[$k]$base[$i]] = $dhd[$index]f;\n";
					printf "param_loop_terminal_S[$base[$l]$base[$j]][$base[$k]$base[$i]] = $entropy" . "f;\n";
					
					printf "\n";
				}
			}
		}
	}
	
	close(DGD);
	close(DHD);
}

main();

exit;
